@@ -15,6 +15,8 @@ from urllib import parse
 from urllib.parse import unquote, urljoin
 
 from pipenv.utils.constants import VCS_LIST
+from pipenv.utils.dependencies import extract_vcs_url
+from pipenv.vendor.tomlkit.items import SingleKey, Table
 
 try:
     import tomllib as toml
@@ -32,7 +34,6 @@ from pipenv.patched.pip._internal.exceptions import ConfigurationError
 from pipenv.patched.pip._internal.models.link import Link
 from pipenv.patched.pip._internal.req.req_install import InstallRequirement
 from pipenv.patched.pip._internal.utils.hashes import FAVORITE_HASH
-from pipenv.patched.pip._vendor import pkg_resources
 from pipenv.utils import err
 from pipenv.utils.constants import is_type_checking
 from pipenv.utils.dependencies import (
@@ -63,13 +64,13 @@ from pipenv.utils.shell import (
     find_windows_executable,
     get_workon_home,
     is_virtual_environment,
-    load_path,
     looks_like_dir,
     safe_expandvars,
     system_which,
 )
 from pipenv.utils.toml import cleanup_toml, convert_toml_outline_tables
-from pipenv.vendor import click, plette, tomlkit
+from pipenv.utils.virtualenv import virtualenv_scripts_dir
+from pipenv.vendor import plette, tomlkit
 
 try:
     # this is only in Python3.8 and later
@@ -78,6 +79,11 @@ except ImportError:
     # eventually distlib will remove cached property when they drop Python3.7
     from pipenv.patched.pip._vendor.distlib.util import cached_property
 
+if sys.version_info < (3, 10):
+    from pipenv.vendor import importlib_metadata
+else:
+    import importlib.metadata as importlib_metadata
+
 if is_type_checking():
     from typing import Dict, List, Union
 
@@ -129,11 +135,6 @@ def preferred_newlines(f):
     return DEFAULT_NEWLINES
 
 
-# (path, file contents) => TOMLFile
-# keeps track of pipfiles that we've seen so we do not need to re-parse 'em
-_pipfile_cache = {}
-
-
 class SourceNotFound(KeyError):
     pass
 
@@ -152,7 +153,7 @@ class Project:
         self._pipfile_newlines = DEFAULT_NEWLINES
         self._lockfile_newlines = DEFAULT_NEWLINES
         self._requirements_location = None
-        self._original_dir = os.path.abspath(os.curdir)
+        self._original_dir = Path.cwd().resolve()
         self._environment = None
         self._build_system = {"requires": ["setuptools", "wheel"]}
         self.python_version = python_version
@@ -213,12 +214,13 @@ class Project:
             with contextlib.suppress(TypeError, AttributeError):
                 os.chdir(self.project_directory)
 
-    def path_to(self, p: str) -> str:
+    def path_to(self, p: str) -> Path:
         """Returns the absolute path to a given relative path."""
-        if os.path.isabs(p):
-            return p
+        path = Path(p)
+        if path.is_absolute():
+            return path
 
-        return os.sep.join([self._original_dir, p])
+        return Path(self._original_dir) / p
 
     def get_pipfile_section(self, section):
         """Returns the details from the section of the Project's Pipfile."""
@@ -236,6 +238,8 @@ class Project:
             return ["packages", "dev-packages"] + list(package_categories)
 
     def get_requests_session_for_source(self, source):
+        if not (source and source.get("name")):
+            return None
         if self.sessions.get(source["name"]):
             session = self.sessions[source["name"]]
         else:
@@ -243,6 +247,7 @@ class Project:
                 self.s.PIPENV_MAX_RETRIES,
                 source.get("verify_ssl", True),
                 cache_dir=self.s.PIPENV_CACHE_DIR,
+                source=source.get("url"),
             )
             self.sessions[source["name"]] = session
         return session
@@ -267,6 +272,8 @@ class Project:
     def get_hashes_from_pypi(self, ireq, source):
         pkg_url = f"https://pypi.org/pypi/{ireq.name}/json"
         session = self.get_requests_session_for_source(source)
+        if not session:
+            return None
         try:
             collected_hashes = set()
             # Grab the hashes from the new warehouse API.
@@ -352,11 +359,8 @@ class Project:
 
         except Exception:
             if self.s.is_verbose():
-                click.echo(
-                    "{}: Error generating hash for {}".format(
-                        click.style("Warning", bold=True, fg="red"), ireq.name
-                    ),
-                    err=True,
+                err.print(
+                    f"[bold red]Warning[/bold red]: Error generating hash for {ireq.name}"
                 )
             return None
 
@@ -374,12 +378,12 @@ class Project:
     @property
     def name(self) -> str:
         if self._name is None:
-            self._name = self.pipfile_location.split(os.sep)[-2]
+            self._name = Path(self.pipfile_location).parent.name
         return self._name
 
     @property
     def pipfile_exists(self) -> bool:
-        return os.path.isfile(self.pipfile_location)
+        return Path(self.pipfile_location).is_file()
 
     @property
     def required_python_version(self) -> str:
@@ -392,7 +396,7 @@ class Project:
 
     @property
     def project_directory(self) -> str:
-        return os.path.abspath(os.path.join(self.pipfile_location, os.pardir))
+        return str(Path(self.pipfile_location).parent.absolute())
 
     @property
     def requirements_exists(self) -> bool:
@@ -402,67 +406,65 @@ class Project:
         if self.s.PIPENV_VENV_IN_PROJECT is False:
             return False
         return self.s.PIPENV_VENV_IN_PROJECT or (
-            self.project_directory
-            and os.path.isdir(os.path.join(self.project_directory, ".venv"))
+            self.project_directory and Path(self.project_directory, ".venv").is_dir()
         )
 
     @property
     def virtualenv_exists(self) -> bool:
-        if os.path.exists(self.virtualenv_location):
-            if os.name == "nt":
-                extra = ["Scripts", "activate.bat"]
+        venv_path = Path(self.virtualenv_location)
+
+        scripts_dir = self.virtualenv_scripts_location
+
+        if venv_path.exists():
+
+            # existence of active.bat is dependent on the platform path prefix
+            # scheme, not platform itself. This handles special cases such as
+            # Cygwin/MinGW identifying as 'nt' platform, yet preferring a
+            # 'posix' path prefix scheme.
+            if scripts_dir.name == "Scripts":
+                activate_path = scripts_dir / "activate.bat"
             else:
-                extra = ["bin", "activate"]
-            return os.path.isfile(os.sep.join([self.virtualenv_location] + extra))
+                activate_path = scripts_dir / "activate"
+            return activate_path.is_file()
 
         return False
 
-    def get_location_for_virtualenv(self) -> str:
+    def get_location_for_virtualenv(self) -> Path:
         # If there's no project yet, set location based on config.
         if not self.project_directory:
             if self.is_venv_in_project():
-                return os.path.abspath(".venv")
-            return str(get_workon_home().joinpath(self.virtualenv_name))
+                return Path(".venv").absolute()
+            return get_workon_home().joinpath(self.virtualenv_name)
 
-        dot_venv = os.path.join(self.project_directory, ".venv")
+        dot_venv = Path(self.project_directory) / ".venv"
 
-        # If there's no .venv in project root, set location based on config.
-        if not os.path.exists(dot_venv):
+        # If there's no .venv in project root or it is a folder, set location based on config.
+        if not dot_venv.exists() or dot_venv.is_dir():
             if self.is_venv_in_project():
                 return dot_venv
-            return str(get_workon_home().joinpath(self.virtualenv_name))
-
-        # If .venv in project root is a directory, use it.
-        if os.path.isdir(dot_venv):
-            return dot_venv
+            return get_workon_home().joinpath(self.virtualenv_name)
 
         # Now we assume .venv in project root is a file. Use its content.
-        with open(dot_venv) as f:
-            name = f.read().strip()
+        name = dot_venv.read_text().strip()
 
         # If .venv file is empty, set location based on config.
         if not name:
-            return str(get_workon_home().joinpath(self.virtualenv_name))
+            return get_workon_home().joinpath(self.virtualenv_name)
 
         # If content looks like a path, use it as a relative path.
         # Otherwise, use directory named after content in WORKON_HOME.
         if looks_like_dir(name):
-            path = Path(self.project_directory, name)
-            return path.absolute().as_posix()
-        return str(get_workon_home().joinpath(name))
-
-    @property
-    def working_set(self) -> pkg_resources.WorkingSet:
-        sys_path = load_path(self.which("python"))
-        return pkg_resources.WorkingSet(sys_path)
+            path = Path(self.project_directory) / name
+            return path.absolute()
+        return get_workon_home().joinpath(name)
 
     @property
     def installed_packages(self):
         return self.environment.get_installed_packages()
 
     @property
-    def installed_package_names(self) -> list[str]:
-        return get_canonical_names([pkg.key for pkg in self.installed_packages])
+    def installed_package_names(self):
+        return get_canonical_names([pkg.name for pkg in self.installed_packages])
 
     @property
     def lockfile_package_names(self) -> dict[str, set[str]]:
@@ -515,7 +517,7 @@ class Project:
             self._environment = self.get_environment(allow_global=allow_global)
         return self._environment
 
-    def get_outdated_packages(self) -> list[pkg_resources.Distribution]:
+    def get_outdated_packages(self) -> list[importlib_metadata.Distribution]:
         return self.environment.get_outdated_packages(pre=self.pipfile.get("pre", False))
 
     @classmethod
@@ -583,8 +585,8 @@ class Project:
         sanitized, encoded_hash = self._get_virtualenv_hash(self.name)
         suffix = ""
         if self.s.PIPENV_PYTHON:
-            if os.path.isabs(self.s.PIPENV_PYTHON):
-                suffix = f"-{os.path.basename(self.s.PIPENV_PYTHON)}"
+            if Path(self.s.PIPENV_PYTHON).is_absolute():
+                suffix = f"-{Path(self.s.PIPENV_PYTHON).name}"
             else:
                 suffix = f"-{self.s.PIPENV_PYTHON}"
 
@@ -601,29 +603,35 @@ class Project:
             and not self.s.PIPENV_IGNORE_VIRTUALENVS
             and virtualenv_env
         ):
-            return virtualenv_env
+            return Path(virtualenv_env)
 
         if not self._virtualenv_location:  # Use cached version, if available.
-            assert self.project_directory, "project not created"
-            self._virtualenv_location = self.get_location_for_virtualenv()
+            if not self.project_directory:
+                raise RuntimeError("Project location not created nor specified")
+            location = self.get_location_for_virtualenv()
+            self._virtualenv_location = Path(location)
         return self._virtualenv_location
 
     @property
-    def virtualenv_src_location(self) -> str:
+    def virtualenv_src_location(self) -> Path:
         if self.virtualenv_location:
-            loc = os.sep.join([self.virtualenv_location, "src"])
+            loc = Path(self.virtualenv_location) / "src"
         else:
-            loc = os.sep.join([self.project_directory, "src"])
-        os.makedirs(loc, exist_ok=True)
+            loc = Path(self.project_directory) / "src"
+        loc.mkdir(parents=True, exist_ok=True)
         return loc
 
     @property
-    def download_location(self) -> str:
+    def virtualenv_scripts_location(self) -> Path:
+        return virtualenv_scripts_dir(self.virtualenv_location)
+
+    @property
+    def download_location(self) -> Path:
         if self._download_location is None:
-            loc = os.sep.join([self.virtualenv_location, "downloads"])
+            loc = Path(self.virtualenv_location) / "downloads"
             self._download_location = loc
         # Create the directory, if it doesn't exist.
-        os.makedirs(self._download_location, exist_ok=True)
+        self._download_location.mkdir(parents=True, exist_ok=True)
         return self._download_location
 
     @property
@@ -672,16 +680,9 @@ class Project:
 
     @property
     def parsed_pipfile(self) -> tomlkit.toml_document.TOMLDocument | TPipfile:
-        """Parse Pipfile into a TOMLFile and cache it
-
-        (call clear_pipfile_cache() afterwards if mutating)"""
+        """Parse Pipfile into a TOMLFile"""
         contents = self.read_pipfile()
-        # use full contents to get around str/bytes 2/3 issues
-        cache_key = (self.pipfile_location, contents)
-        if cache_key not in _pipfile_cache:
-            parsed = self._parse_pipfile(contents)
-            _pipfile_cache[cache_key] = parsed
-        return _pipfile_cache[cache_key]
+        return self._parse_pipfile(contents)
 
     def read_pipfile(self) -> str:
         # Open the pipfile, read it into memory.
@@ -693,10 +694,6 @@ class Project:
 
         return contents
 
-    def clear_pipfile_cache(self) -> None:
-        """Clear pipfile cache (e.g., so we can mutate parsed pipfile)"""
-        _pipfile_cache.clear()
-
     def _parse_pipfile(
         self, contents: str
     ) -> tomlkit.toml_document.TOMLDocument | TPipfile:
@@ -708,11 +705,12 @@ class Project:
             return toml.loads(contents)
 
     def _read_pyproject(self) -> None:
-        pyproject = self.path_to("pyproject.toml")
-        if os.path.exists(pyproject):
-            self._pyproject = toml.load(pyproject)
+        pyproject_path = Path(self.path_to("pyproject.toml"))
+        if pyproject_path.exists():
+            self._pyproject = toml.load(pyproject_path)
             build_system = self._pyproject.get("build-system", None)
-            if not os.path.exists(self.path_to("setup.py")):
+            setup_py_path = Path(self.path_to("setup.py"))
+            if not setup_py_path.exists():
                 if not build_system or not build_system.get("requires"):
                     build_system = {
                         "requires": ["setuptools>=40.8.0", "wheel"],
@@ -751,7 +749,7 @@ class Project:
     def update_settings(self, d: dict[str, str | bool]) -> None:
         settings = self.settings
         changed = False
-        for new in d:
+        for new in d.keys():  # noqa: PLC0206
             if new not in settings:
                 settings[new] = d[new]
                 changed = True
@@ -782,12 +780,7 @@ class Project:
         for category in categories:
             lock_section = lockfile.get(category)
             if lock_section is None:
-                lockfile[category] = lock_section = {}
-            for key in list(lock_section.keys()):
-                norm_key = pep423_name(key)
-                specifier = lock_section[key]
-                del lock_section[key]
-                lockfile[category][norm_key] = specifier
+                lockfile[category] = {}
 
         return lockfile
 
@@ -804,7 +797,7 @@ class Project:
 
     @property
     def lockfile_exists(self):
-        return os.path.isfile(self.lockfile_location)
+        return Path(self.lockfile_location).is_file()
 
     @property
     def lockfile_content(self):
@@ -880,10 +873,7 @@ class Project:
         # Default requires.
         required_python = python
         if not python:
-            if self.virtualenv_location:
-                required_python = self.which("python", self.virtualenv_location)
-            else:
-                required_python = self.which("python")
+            required_python = self.which("python")
         version = python_version(required_python) or self.s.PIPENV_DEFAULT_PYTHON_VERSION
         if version:
             data["requires"] = {"python_version": ".".join(version.split(".")[:2])}
@@ -993,8 +983,6 @@ class Project:
         formatted_data = cleanup_toml(formatted_data)
         with open(path, "w", newline=newlines) as f:
             f.write(formatted_data)
-        # pipfile is mutated!
-        self.clear_pipfile_cache()
 
     def write_lockfile(self, content):
         """Write out the lockfile."""
@@ -1013,9 +1001,7 @@ class Project:
             if os.environ.get("PIPENV_PYPI_MIRROR"):
                 sources[0]["url"] = os.environ["PIPENV_PYPI_MIRROR"]
             return sources
-        # We need to make copies of the source info so we don't
-        # accidentally modify the cache. See #2100 where values are
-        # written after the os.path.expandvars() call.
+        # Expand environment variables in the source URLs.
         sources = [
             {k: safe_expandvars(v) if expand_vars else v for k, v in source.items()}
             for source in self.parsed_pipfile["source"]
@@ -1026,13 +1012,18 @@ class Project:
         return sources
 
     def get_default_index(self):
-        return self.pipfile_sources()[0]
+        return self.populate_source(self.pipfile_sources()[0])
 
     def get_index_by_name(self, index_name):
         for source in self.pipfile_sources():
             if source.get("name") == index_name:
                 return source
 
+    def get_index_by_url(self, index_url):
+        for source in self.pipfile_sources():
+            if source.get("url") == index_url:
+                return source
+
     @property
     def sources(self):
         if self.lockfile_exists and hasattr(self.lockfile_content, "keys"):
@@ -1090,7 +1081,6 @@ class Project:
 
         sources = (self.sources, self.pipfile_sources())
         if refresh:
-            self.clear_pipfile_cache()
             sources = reversed(sources)
         found = next(
             iter(find_source(source, name=name, url=url) for source in sources), None
@@ -1101,15 +1091,36 @@ class Project:
         return found
 
     def get_package_name_in_pipfile(self, package_name, category):
-        """Get the equivalent package name in pipfile"""
-        section = self.parsed_pipfile.get(category)
-        if section is None:
-            section = {}
-        package_name = pep423_name(package_name)
+        section = self.parsed_pipfile.get(category, {})
+        normalized_name = pep423_name(package_name)
         for name in section:
-            if pep423_name(name) == package_name:
+            if pep423_name(name) == normalized_name:
                 return name
-        return None
+        return package_name  # Return original name if not found
+
+    def get_pipfile_entry(self, package_name, category):
+        name = self.get_package_name_in_pipfile(package_name, category)
+        return self.parsed_pipfile.get(category, {}).get(name)
+
+    def _sort_category(self, category) -> Table:
+        # copy table or create table from dict-like object
+        table = tomlkit.table()
+        if isinstance(category, Table):
+            table.update(category.value)
+        else:
+            table.update(category)
+
+        # sort the table internally
+        table._value._body.sort(key=lambda t: t[0] and t[0].key or "")
+        for index, (key, _) in enumerate(table._value._body):
+            assert isinstance(key, SingleKey)
+            indices = table._value._map[key]
+            if isinstance(indices, tuple):
+                table._value._map[key] = (index,) + indices[1:]
+            else:
+                table._value._map[key] = index
+
+        return table
 
     def remove_package_from_pipfile(self, package_name, category):
         # Read and append Pipfile.
@@ -1117,6 +1128,18 @@ class Project:
         p = self.parsed_pipfile
         if name:
             del p[category][name]
+            if self.settings.get("sort_pipfile"):
+                p[category] = self._sort_category(p[category])
+            self.write_toml(p)
+            return True
+        return False
+
+    def reset_category_in_pipfile(self, category):
+        # Read and append Pipfile.
+        p = self.parsed_pipfile
+        if category:
+            del p[category]
+            p[category] = {}
             self.write_toml(p)
             return True
         return False
@@ -1133,7 +1156,9 @@ class Project:
                 del parsed[category][pkg_name]
         self.write_toml(parsed)
 
-    def generate_package_pipfile_entry(self, package, pip_line, category=None):
+    def generate_package_pipfile_entry(
+        self, package, pip_line, category=None, index_name=None
+    ):
         """Generate a package entry from pip install line
         given the installreq package and the pip line that generated it.
         """
@@ -1181,7 +1206,9 @@ class Project:
                     vcs_parts = vcs_part.rsplit("@", 1)
                     if len(vcs_parts) > 1:
                         entry["ref"] = vcs_parts[1].split("#", 1)[0].strip()
-                    entry[vcs] = vcs_parts[0].strip()
+                    vcs_url = vcs_parts[0].strip()
+                    vcs_url = extract_vcs_url(vcs_url)
+                    entry[vcs] = vcs_url
 
                     # Check and extract subdirectory fragment
                     if package.link.subdirectory_fragment:
@@ -1189,7 +1216,10 @@ class Project:
                     break
         else:
             entry["version"] = specifier
-        if hasattr(package, "index"):
+
+        if index_name:
+            entry["index"] = index_name
+        elif hasattr(package, "index"):
             entry["index"] = package.index
 
         if len(entry) == 1 and "version" in entry:
@@ -1212,23 +1242,28 @@ class Project:
         newly_added = False
 
         # Read and append Pipfile.
-        p = self.parsed_pipfile
+        parsed_pipfile = self.parsed_pipfile
 
         # Set empty group if it doesn't exist yet.
-        if category not in p:
-            p[category] = {}
+        if category not in parsed_pipfile:
+            parsed_pipfile[category] = {}
 
-        if name and name != normalized_name:
-            self.remove_package_from_pipfile(name, category=category)
+        section = parsed_pipfile.get(category, {})
+        for entry_name in section.copy().keys():
+            if entry_name.lower() == normalized_name.lower():
+                del parsed_pipfile[category][entry_name]
 
         # Add the package to the group.
-        if normalized_name not in p[category]:
+        if normalized_name not in parsed_pipfile[category]:
             newly_added = True
 
-        p[category][normalized_name] = entry
+        parsed_pipfile[category][normalized_name] = entry
+
+        if self.settings.get("sort_pipfile"):
+            parsed_pipfile[category] = self._sort_category(parsed_pipfile[category])
 
         # Write Pipfile.
-        self.write_toml(p)
+        self.write_toml(parsed_pipfile)
         return newly_added, category, normalized_name
 
     def src_name_from_url(self, index_url):
@@ -1249,26 +1284,45 @@ class Project:
         return name
 
     def add_index_to_pipfile(self, index, verify_ssl=True):
-        """Adds a given index to the Pipfile."""
+        """
+        Adds a given index to the Pipfile if it doesn't already exist.
+        Returns the source name regardless of whether it was newly added or already existed.
+        """
         # Read and append Pipfile.
         p = self.parsed_pipfile
         source = None
+
+        # Try to find existing source by URL or name
         try:
             source = self.get_source(url=index)
         except SourceNotFound:
             with contextlib.suppress(SourceNotFound):
                 source = self.get_source(name=index)
 
+        # If we found an existing source with a name, return it
         if source is not None and source.get("name"):
             return source["name"]
-        source = {"url": index, "verify_ssl": verify_ssl}
-        source["name"] = self.src_name_from_url(index)
-        # Add the package to the group.
+
+        # Check if the URL already exists in any source
+        if "source" in p:
+            for existing_source in p["source"]:
+                if existing_source.get("url") == index:
+                    return existing_source.get("name")
+
+        # If we reach here, the source doesn't exist, so create and add it
+        source = {
+            "url": index,
+            "verify_ssl": verify_ssl,
+            "name": self.src_name_from_url(index),
+        }
+
+        # Add the source to the group
         if "source" not in p:
             p["source"] = [tomlkit.item(source)]
         else:
             p["source"].append(tomlkit.item(source))
-        # Write Pipfile.
+
+        # Write Pipfile
         self.write_toml(p)
         return source["name"]
 
@@ -1278,28 +1332,34 @@ class Project:
 
     def load_lockfile(self, expand_env_vars=True):
         lockfile_modified = False
-        with open(self.lockfile_location, encoding="utf-8") as lock:
-            try:
-                j = json.load(lock)
-                self._lockfile_newlines = preferred_newlines(lock)
-            except JSONDecodeError:
-                click.secho(
-                    "Pipfile.lock is corrupted; ignoring contents.",
-                    fg="yellow",
-                    bold=True,
-                    err=True,
-                )
-                j = {}
+        lockfile_path = Path(self.lockfile_location)
+        pipfile_path = Path(self.pipfile_location)
+
+        try:
+            with lockfile_path.open(encoding="utf-8") as lock:
+                try:
+                    j = json.load(lock)
+                    self._lockfile_newlines = preferred_newlines(lock)
+                except JSONDecodeError:
+                    err.print(
+                        "[bold yellow]Pipfile.lock is corrupted; ignoring contents.[/bold yellow]"
+                    )
+                    j = {}
+        except FileNotFoundError:
+            j = {}
+
         if not j.get("_meta"):
-            with open(self.pipfile_location) as pf:
+            with pipfile_path.open() as pf:
                 default_lockfile = plette.Lockfile.with_meta_from(
                     plette.Pipfile.load(pf), categories=[]
                 )
                 j["_meta"] = default_lockfile._data["_meta"]
                 lockfile_modified = True
+
         if j.get("default") is None:
             j["default"] = {}
             lockfile_modified = True
+
         if j.get("develop") is None:
             j["develop"] = {}
             lockfile_modified = True
@@ -1310,6 +1370,7 @@ class Project:
         if expand_env_vars:
             # Expand environment variables in Pipfile.lock at runtime.
             for i, _ in enumerate(j["_meta"].get("sources", {})):
+                # Path doesn't have expandvars method, so we need to use os.path.expandvars
                 j["_meta"]["sources"][i]["url"] = os.path.expandvars(
                     j["_meta"]["sources"][i]["url"]
                 )
@@ -1317,7 +1378,8 @@ class Project:
         return j
 
     def get_lockfile_hash(self):
-        if not os.path.exists(self.lockfile_location):
+        lockfile_path = Path(self.lockfile_location)
+        if not lockfile_path.exists():
             return
 
         try:
@@ -1373,10 +1435,10 @@ class Project:
     def finders(self):
         from .vendor.pythonfinder import Finder
 
-        scripts_dirname = "Scripts" if os.name == "nt" else "bin"
-        scripts_dir = os.path.join(self.virtualenv_location, scripts_dirname)
         finders = [
-            Finder(path=scripts_dir, global_search=gs, system=False)
+            Finder(
+                path=str(self.virtualenv_scripts_location), global_search=gs, system=False
+            )
             for gs in (False, True)
         ]
         return finders
@@ -1385,14 +1447,11 @@ class Project:
     def finder(self):
         return next(iter(self.finders), None)
 
-    def which(self, search, as_path=True):
+    def which(self, search):
         find = operator.methodcaller("which", search)
         result = next(iter(filter(None, (find(finder) for finder in self.finders))), None)
         if not result:
             result = self._which(search)
-        else:
-            if as_path:
-                result = str(result.path)
         return result
 
     def python(self, system=False) -> str:
@@ -1407,22 +1466,37 @@ class Project:
                 location = self.virtualenv_location
             else:
                 location = os.environ.get("VIRTUAL_ENV", None)
-        if not (location and os.path.exists(location)) and not allow_global:
+
+        location_path = Path(location) if location else None
+
+        if not (location_path and location_path.exists()) and not allow_global:
             raise RuntimeError("location not created nor specified")
 
-        version_str = "python{}".format(".".join([str(v) for v in sys.version_info[:2]]))
-        is_python = command in ("python", os.path.basename(sys.executable), version_str)
+        version_str = f"python{'.'.join([str(v) for v in sys.version_info[:2]])}"
+        is_python = command in ("python", Path(sys.executable).name, version_str)
+
         if not allow_global:
+            scripts_location = virtualenv_scripts_dir(location_path)
+
             if os.name == "nt":
-                p = find_windows_executable(os.path.join(location, "Scripts"), command)
+                p = find_windows_executable(str(scripts_location), command)
+                # Convert to Path object if it's a string
+                p = Path(p) if isinstance(p, str) else p
             else:
-                p = os.path.join(location, "bin", command)
+                p = scripts_location / command
+        elif is_python:
+            p = Path(sys.executable)
         else:
+            p = None
+
+        if p is None or not p.exists():
             if is_python:
-                p = sys.executable
-        if not os.path.exists(p):
-            if is_python:
-                p = sys.executable or system_which("python")
+                p = (
+                    Path(sys.executable)
+                    if sys.executable
+                    else Path(system_which("python"))
+                )
             else:
-                p = system_which(command)
+                p = Path(system_which(command)) if system_which(command) else None
+
         return p
